from avalonBG.rules import get_rules


def test_get_rules():

    rules = get_rules()

    for nb_player, rules_player in rules.items():
        assert 5 <= int(nb_player) <= 10
        assert set(rules_player) == {"blue", "red", "quests"}



from random import choice, sample

import requests as req


def new_vote(game_id, url, result=None):

    game = req.get("{}games/{}".format(url, game_id)).json()
    list_players_quest = sample(game["players"], game["quests"][game["current_quest"]]["nb_players_to_send"])

    req.put(
        "{}games/{}/quests/{}".format(url, game_id, game["current_quest"]),
        json=[dictio["id"] for dictio in list_players_quest]
    )

    nb_votes = 0
    if result is not None:
        nb_votes = game["quests"][game["current_quest"]]["nb_votes_to_fail"]
        if result:
            nb_votes = len(list_players_quest) - (game["quests"][game["current_quest"]]["nb_votes_to_fail"] - 1)

    for ind, player in enumerate(list_players_quest):

        if ind < nb_votes:
            req.post(
                "{}games/{}/quests/{}".format(url, game_id, game["current_quest"]),
                json={player["id"]: result}
            )

        else:
            req.post(
                "{}games/{}/quests/{}".format(url, game_id, game["current_quest"]),
                json={player["id"]: choice([True, True, False])}
            )

    req.get("{}games/{}".format(url, game_id))


def quests(game_id, url, result=None):

    while "result" not in req.get("{}games/{}".format(url, game_id)).json():
        new_vote(game_id, url, result)
        req.get("{}games/{}".format(url, game_id))


def complete_game(roles, nb_players=choice(range(5, 11)), vote_assassin=None, result=None, url="http://localhost:5000/"):

    put_games = {"names": ["player{}".format(nb+1) for nb in range(nb_players)], "roles": roles}
    req_put_game = req.put("{}games".format(url), json=put_games)

    game_id = req_put_game.json()["id"]

    # req.get("{}games/{}/mp3".format(url, game_id))

    quests(game_id=game_id, url=url, result=result)

    req.get("{}games/{}".format(url, game_id))

    if vote_assassin is False:
        assassin_id = [dico for dico in req_put_game.json()["players"] if "assassin" in dico][0]["id"]
        merlin = [dico for dico in req_put_game.json()["players"] if dico["role"] == "merlin"][0]["id"]
        req.post("{}games/{}/guess_merlin".format(url, game_id), json={assassin_id: merlin})

    req.get("{}games/{}".format(url, game_id))


def main():

    # put_restart_db = ["games", "players", "quests", "users"]
    # req.put("{}restart_db".format(url), json=put_restart_db)

    complete_game(roles=[])


if __name__ == '__main__':

    main()



import rethinkdb as r

from avalonBG.db_utils import db_get_value, db_update_value, db_get_game
from avalonBG.exception import AvalonBGError


def check_quest_number(quest_number):
    if not 0 <= quest_number <= 4:
        raise AvalonBGError("Quest's number should be between 0 and 4!")


def update_current_id_player(game_id):
    """This function update 'current_id_player' of the game game_id."""

    list_id_players = db_get_value(
        table="games",
        ident=game_id,
        key="players"
    )

    list_current_id_player = db_get_value(
        table="games",
        ident=game_id,
        key="current_id_player"
    )

    current_ind_player = list_id_players.index(list_current_id_player)
    next_ind_player = (current_ind_player + 1) % len(list_id_players)
    db_update_value(
        table="games",
        ident=game_id,
        key="current_id_player",
        value=list_id_players[next_ind_player]
    )


def quest_unsend(game_id):
    """This function sends new quest of the game <game_id>"""

    nb_quest_unsend = db_get_value(
        table="games",
        ident=game_id,
        key="nb_quest_unsend"
    )

    if nb_quest_unsend == 5:
        raise AvalonBGError("Game is over because 5 consecutive laps have been passed: Red team won!")

    if nb_quest_unsend < 4:
        update_current_id_player(game_id=game_id)
        value = 1 + db_get_value(
            table="games",
            ident=game_id,
            key="nb_quest_unsend"
        )
        db_update_value(
            table="games",
            ident=game_id,
            key="nb_quest_unsend",
            value=value
        )

    if nb_quest_unsend == 4:
        value = 1 + db_get_value(
            table="games",
            ident=game_id,
            key="nb_quest_unsend"
        )
        db_update_value(
            table="games",
            ident=game_id,
            key="nb_quest_unsend",
            value=value
        )
        r.RethinkDB().table("games").get(game_id).update(
            {
                "result": {
                    "status": False
                }
            },
            return_changes=True
        )["changes"][0]["new_val"].run()

    game_updated = db_get_game(game_id=game_id)

    return game_updated


def quest_delete(game_id, quest_number):

    check_quest_number(quest_number=quest_number)

    id_quest_number = db_get_value(
        table="games",
        ident=game_id,
        key="quests"
    )[quest_number]

    game_updated = r.RethinkDB().table("quests").get(id_quest_number).replace(
        r.RethinkDB().row.without("votes", "status"),
        return_changes=True
    )["changes"][0]["new_val"].run()

    return game_updated


def quest_get(game_id, quest_number):

    check_quest_number(quest_number=quest_number)

    game = r.RethinkDB().table("games").get(game_id).run()
    if not game:
        raise AvalonBGError("Game's id '{}' does not exist!".format(game_id))

    game_updated = r.RethinkDB().table("quests").get(game["quests"][quest_number]).run()

    if "status" not in game_updated:
        raise AvalonBGError("The vote number '{}' has not started!".format(quest_number))

    if game_updated["status"] is None:
        raise AvalonBGError("The vote number '{}' is not finished!".format(quest_number))

    return game_updated


def quest_post(payload, game_id, quest_number):

    check_quest_number(quest_number=quest_number)

    game = r.RethinkDB().table("games").get(game_id).run()
    if not game:
        raise AvalonBGError("Game's id '{}' does not exist!".format(game_id))

    if game["nb_quest_unsend"] == 5:
        raise AvalonBGError("Game is over because 5 consecutive laps have been passed: Red team won!")

    if "result" in game:
        raise AvalonBGError("Game is over!")

    if game["current_quest"] != quest_number:
        raise AvalonBGError("Only vote number {} is allowed!".format(game["current_quest"]))

    quest = r.RethinkDB().table("quests").get(game["quests"][quest_number]).run()

    if "status" not in quest:
        raise AvalonBGError("Vote number '{}' is not established!".format(quest_number))

    if quest["status"] is not None:
        raise AvalonBGError("Vote number '{}' is finished!".format(quest_number))

    if len(payload) != 1:
        raise AvalonBGError("Only one vote allowed!")

    if list(payload)[0] not in list(quest["votes"]):
        raise AvalonBGError("Player '{}' is not allowed to vote!".format(list(payload)[0]))

    if quest["votes"][list(payload)[0]] is not None:
        raise AvalonBGError("Player '{}' has already voted!".format(list(payload)[0]))

    if not isinstance(list(payload.values())[0], bool):
        raise AvalonBGError("Vote should be a boolean!")

    quest["votes"].update(payload)

    list_vote = list(quest["votes"].values())
    if not list_vote.count(None):
        quest["status"] = not list_vote.count(False) >= quest["nb_votes_to_fail"]
        game["current_id_player"] = game["players"][
            (game["players"].index(game["current_id_player"]) + 1) % len(game["players"])
        ]
        game["nb_quest_unsend"] = 0
        game["current_quest"] = game["current_quest"] + 1

    if len(quest["votes"]) != quest["nb_players_to_send"]:
        r.RethinkDB().table("quests").get(game["quests"][quest_number]).update({"votes": None}).run()

    game_updated = r.RethinkDB().table("quests").get(game["quests"][quest_number]).replace(
        quest,
        return_changes=True
    )["changes"][0]["new_val"].run()

    list_status = [
        quest.get("status") for quest in r.RethinkDB().table("quests").get_all(r.RethinkDB().args(game["quests"])).run()
    ]

    if list_status.count(False) >= 3:
        game["result"] = {"status": False}

    if list_status.count(True) >= 3:
        game["result"] = {"status": True}

    if not list_vote.count(None):
        r.RethinkDB().table("games").get(game_id).replace(game).run()

    return game_updated


def quest_put(payload, game_id, quest_number):

    check_quest_number(quest_number=quest_number)

    game = r.RethinkDB().table("games").get(game_id).run()
    if not game:
        raise AvalonBGError("Game's id '{}'' does not exist!".format(game_id))

    if game["nb_quest_unsend"] == 5:
        raise AvalonBGError("Game is over because 5 consecutive laps have been passed : Red team won!")

    if "result" in game:
        raise AvalonBGError("Game is over!")

    if game["current_quest"] != quest_number:
        raise AvalonBGError("Vote number {} is already established!".format(quest_number))

    quest = r.RethinkDB().table("quests").get(game["quests"][quest_number]).run()

    if ("status" in quest or "votes" in quest) and quest["status"] is not None:
        raise AvalonBGError("Only vote number '{}'' is allowed!".format(game["current_quest"]))

    for player_id in payload:
        if player_id not in game["players"]:
            raise AvalonBGError("Player '{}' is not in this game!".format(player_id))

    id_quest_number = game["quests"][quest_number]
    nb_players_to_send = db_get_value(
        table="quests",
        ident=id_quest_number,
        key="nb_players_to_send"
    )

    if len(payload) != nb_players_to_send:
        raise AvalonBGError("Quest number '{}' needs '{}' votes!".format(quest_number, nb_players_to_send))

    if "status" in quest:
        r.RethinkDB().table("quests").get(id_quest_number).update({"votes": None}).run()

    game_updated = (
        r.RethinkDB().table("quests").get(id_quest_number).update(
            {
                "votes": {player_id: None for player_id in payload},
                "status": None
            },
            return_changes=True
        )["changes"][0]["new_val"].run()
    )

    return game_updated



"""File that contains exception class used in the package avalonBG"""


class AvalonBGError(Exception):
    """Class AvalonBGError related to specific exception"""

    def __init__(self, msg, log_type="error"):
        Exception.__init__(self, msg)



from random import shuffle, choice

import rethinkdb as r

from avalonBG.db_utils import db_get_value, resolve_key_id
from avalonBG.exception import AvalonBGError
from avalonBG.rules import get_rules


def game_put(payload):
    """
    This function adds rules and roles to players randomly.
        - method: PUT
        - route: /games
        - payload example: {
                               "players": [
                                    {
                                        "name": "name0",
                                        "avatar_index": 0
                                    },
                                    {
                                        "name": "name1",
                                        "avatar_index": 1
                                    },
                                    {
                                        "name": "name2",
                                        "avatar_index": 2
                                    },
                                    {
                                        "name": "name3",
                                        "avatar_index": 3
                                    },
                                    {
                                        "name": "name4",
                                        "avatar_index": 4
                                    }
                               ],
                               "roles": [
                                   "oberon",
                                   "perceval",
                                   "morgan"
                               ]
                           }
        - response example: {
                                "id": "2669a9fe-37c4-4139-ab78-8e3f0d0607d0",
                                "players": [
                                    {
                                        "avatar_index": 0,
                                        "id": "95763b27-de50-4d39-8ac2-2a7010281788",
                                        "name": "name0",
                                        "role": "assassin",
                                        "team": "red"
                                    },
                                    {
                                        ...
                                    },
                                    {
                                        "avatar_index": 4,
                                        "id": "83d21d25-f359-4ddc-9048-69ba1e6cf5b5",
                                        "name": "name4",
                                        "role": "morgan",
                                        "team": "red"
                                    }
                                ]
                            }
    """
    if not payload or set(payload) != {"players", "roles"}:
        raise AvalonBGError("Only 'players' and 'roles' are required!")

    for key, val in payload.items():
        if not isinstance(val, list):
            raise AvalonBGError("Both 'players' and 'roles' must be a list!")

        if key == "players":
            for player in val:
                if not isinstance(player, dict):
                    raise AvalonBGError("'player' must be an object!")

                if set(player) != {"name", "avatar_index"}:
                    raise AvalonBGError("'Only 'name' and 'avatar_index' are required for a player!")

                if not isinstance(player["name"], str):
                    raise AvalonBGError("'name' of a player must be a string!")

                if not isinstance(player["avatar_index"], int):
                    raise AvalonBGError("'avatar_index' of a player must be an int!")

        elif key == "roles":
            for elem in val:
                if not isinstance(elem, str):
                    raise AvalonBGError("Elements in 'roles' must be a string!")

    rules = get_rules()
    min_nb_player = int(min(rules, key=int))
    max_nb_player = int(max(rules, key=int))

    if not min_nb_player <= len(payload["players"]) <= max_nb_player:
        raise AvalonBGError("Player number should be between {} and {}!".format(min_nb_player, max_nb_player))

    if len(payload["players"]) != len(set([player["name"] for player in payload["players"]])):
        raise AvalonBGError("Players name should be unique!")

    for player in payload["players"]:
        if player["name"].isspace() or player["name"] == "":
            raise AvalonBGError("Players' name cannot be empty!")

    if len(payload["roles"]) != len(set(payload["roles"])):
        raise AvalonBGError("Players role should be unique!")

    available_roles = ("oberon", "morgan", "mordred", "perceval")  #TODO: should be a variable
    for role in payload["roles"]:
        if role not in available_roles:
            raise AvalonBGError("Players role should be {}, {}, {} or {}!".format(*available_roles))

    if "morgan" in payload["roles"] and "perceval" not in payload["roles"]:
        raise AvalonBGError("'morgan' is selected but 'perceval' is not!")

    if "perceval" in payload["roles"] and "morgan" not in payload["roles"]:
        raise AvalonBGError("'perceval' is selected but 'morgan' is not!")

    # find rules
    game_rules = rules[str(len(payload["players"]))]

    if len([role for role in payload["roles"] if role != "perceval"]) > game_rules["red"]:
        raise AvalonBGError("Too many red roles chosen!")

    # add roles to players
    players = roles_and_players(
        dict_names_roles=payload,
        max_red=game_rules["red"],
        max_blue=game_rules["blue"]
    )

    # find players
    list_id_players = r.RethinkDB().table("players").insert(players).run()["generated_keys"]

    # find quests
    list_id_quests = r.RethinkDB().table("quests").insert(game_rules["quests"]).run()["generated_keys"]

    inserted_game = r.RethinkDB().table("games").insert(
        {
            "players": list_id_players,
            "quests": list_id_quests,
            "current_id_player": list_id_players[choice(range(len(payload["players"])))],
            "current_quest": 0,
            "nb_quest_unsend": 0
        },
        return_changes=True
    )["changes"][0]["new_val"].run()

    inserted_game.update(
        {
            "players": resolve_key_id(table="players", list_id=inserted_game["players"]),
            "quests": resolve_key_id(table="quests", list_id=inserted_game["quests"])
        }
    )

    return inserted_game


def roles_and_players(dict_names_roles, max_red, max_blue):
    """Check the validity of proposed roles.
    cases break rules: - 1. morgan in the game but Perceval is not
                       - 2. perceval in the game but Morgan is not
                       - 3. Unvalid role
                       - 4. Too many red in the game (or too many blue in the game, checked but impossible)"""

    nb_red, nb_blue = 0, 1
    list_roles = ["merlin"]
    for role in dict_names_roles["roles"]:
        if role in ["mordred", "morgan", "oberon"]:
            nb_red += 1
            list_roles.append(role)
        elif role == "perceval":
            nb_blue += 1
            list_roles.append(role)

    list_roles.extend(["red"]*(max_red-nb_red))
    list_roles.extend(["blue"]*(max_blue-nb_blue))

    shuffle(list_roles)

    bool_assassin = True
    list_players = []
    for ind, role in enumerate(list_roles):
        player = {
            "avatar_index": dict_names_roles["players"][ind]["avatar_index"],
            "name": dict_names_roles["players"][ind]["name"],
            "role": role,
            "team": "blue"
        }

        if role not in ("merlin", "perceval", "blue"):
            player["team"] = "red"
            if bool_assassin:
                bool_assassin = False
                player["assassin"] = True

        list_players.append(player)

    return list_players


def game_guess_merlin(game_id, payload):

    if len(payload) != 1:
        raise AvalonBGError("Only 1 vote required ('assassin')!")

    game = r.RethinkDB().table("games").get(game_id).run()
    if not game:
        raise AvalonBGError("Game's id {} does not exist!".format(game_id))

    if game["nb_quest_unsend"] == 5:
        raise AvalonBGError("Game is over because 5 consecutive laps have been passed : Red team won!")

    assassin_id = list(payload)[0]
    vote_assassin = payload[assassin_id]

    if assassin_id not in game["players"]:
        raise AvalonBGError("Player {} is not in this game!".format(assassin_id))

    if "assassin" not in r.RethinkDB().table("players").get(assassin_id).run():
        raise AvalonBGError("Player {} is not 'assassin'!".format(assassin_id))

    if vote_assassin not in game["players"]:
        raise AvalonBGError("Player {} is not in this game!".format(vote_assassin))

    game = r.RethinkDB().table("games").get(game_id).run()
    if not game:
        raise AvalonBGError("Game's id {} does not exist!".format(game_id))

    result = game.get("result")
    if not result:
        raise AvalonBGError("Game's status is not established!")

    if not result["status"]:
        raise AvalonBGError("Games's status should be 'true' (ie blue team won)!")

    if "guess_merlin_id" in result:
        raise AvalonBGError("Merlin already chosen!")

    result["guess_merlin_id"] = vote_assassin
    if db_get_value("players", vote_assassin, "role") == "merlin":
        result["status"] = False

    updated_game = r.RethinkDB().table("games").get(game_id).update(
        {"result": result},
        return_changes=True
    )["changes"][0]["new_val"].run()

    updated_game.update(
        {
            "players": resolve_key_id(table="players", list_id=updated_game["players"]),
            "quests": resolve_key_id(table="quests", list_id=updated_game["quests"])
        }
    )

    return updated_game



import rethinkdb as r

from avalonBG.exception import AvalonBGError


def db_connect():
    """This function opens the connection to the database."""

    r.RethinkDB().connect("rethinkdb", 28015).repl()
    # r.RethinkDB().connect("0.0.0.0", 28015).repl()


def resolve_key_id(table, list_id):

    list_players = list(r.RethinkDB().table(table).get_all(*list_id).run())
    dict_players = {dict_player["id"]: dict_player for dict_player in list_players}

    return [dict_players[ident] for ident in list_id]


def db_get_game(game_id=None):
    """This function visualize the game of the <game_id>"""

    if not game_id:
        return db_get_table(table_name="games")

    game = r.RethinkDB().table("games").get(game_id).run()
    if not game:
        raise AvalonBGError("Game's id {} does not exist !".format(game_id))

    game.update(
        {
            "players": resolve_key_id(table="players", list_id=game["players"]),
            "quests": resolve_key_id(table="quests", list_id=game["quests"])
        }
    )

    return game


def db_get_table(table_name):
    """This function returns the table associated to the table name."""
    return list(r.RethinkDB().table(table_name).run())


def db_get_value(table, ident, key):
    """This function finds the key value in the table."""
    try:
        requested_value = r.RethinkDB().table(table).get(ident)[key].run()
    except r.errors.ReqlNonExistenceError:
        raise AvalonBGError("Requested value is not available!")

    return requested_value


def db_update_value(table, ident, key, value):
    """This function updates the key value in the table."""

    return r.RethinkDB().table(table).get(ident).update({key: value}).run()


def restart_db(payload_tables):

    if not isinstance(payload_tables, list):
        raise AvalonBGError("Payload should be a list!")

    list_tables = ("games", "players", "quests", "users")
    for table in payload_tables:
        if table not in list_tables:
            raise AvalonBGError("Table {} should be in {}!".format(table, list_tables))

        if table in r.RethinkDB().db("test").table_list().run():
            r.RethinkDB().table_drop(table).run()

        # initialize table
        r.RethinkDB().table_create(table).run()

    return ""



import json

import importlib.resources as pkg_resources

from avalonBG.exception import AvalonBGError

from . import resources


def get_rules():

    filename = "rules.json"

    try:
        with pkg_resources.path(resources, filename) as rules_path:
            with open(rules_path, "r") as infile:
                rules = json.load(infile)
    except FileNotFoundError:
        raise AvalonBGError("File '{}' doesn't exist!".format(filename))
    except json.decoder.JSONDecodeError:
        raise AvalonBGError("File '{}' can't be decoded!".format(filename))

    return rules



import importlib.resources as pkg_resources
from pathlib import Path

from pydub import AudioSegment

from avalonBG.db_utils import db_get_value
from avalonBG.exception import AvalonBGError
from .resources import audio


def create_mp3(output_mp3_path="resources"):
    """Create mp3 all files depending on roles in the game"""

    Path(output_mp3_path).mkdir(parents=True, exist_ok=True)

    #TODO: les roles ne devraient pas etre en dur
    list_all_roles = [
        [],
        ["morgan"],
        ["oberon"],
        ["mordred"],
        ["morgan", "oberon"],
        ["morgan", "mordred"],
        ["oberon", "mordred"],
        ["morgan", "oberon", "mordred"]
    ]

    for list_roles in list_all_roles:

        list_mp3 = ["init.mp3", "serv_mord.mp3"]
        if "oberon" in list_roles:
            list_mp3.append("oberon.mp3")
        list_mp3.append("red_identi.mp3")

        if "morgan" in list_roles:
            list_mp3.append("add_per_mor.mp3")

        list_mp3.append("serv_mord.mp3")
        if "mordred" in list_roles:
            list_mp3.append("mordred.mp3")
        list_mp3.extend(["merlin_identi.mp3", "end.mp3"])

        mp3_combined = AudioSegment.empty()
        for mp3 in list_mp3:
            with pkg_resources.path(audio, mp3) as mp3_path:
                mp3_combined += AudioSegment.from_mp3(mp3_path)

        mp3_path_file = Path(output_mp3_path, "_{}.mp3".format('-'.join(sorted(list_roles)))).as_posix()
        mp3_combined.export(mp3_path_file, format="mp3")


def get_mp3_roles_path(game_id, output_mp3_path="resources"):

    # find role of each player
    list_roles = [
        db_get_value("players", player_id, "role") for player_id in db_get_value("games", game_id, "players") \
        if db_get_value("players", player_id, "role") not in ("blue", "merlin", "perceval", "red")
    ]

    name_roles = "-".join(sorted(list_roles))
    mp3_roles_path = Path(output_mp3_path, "_{}.mp3".format(name_roles))

    if not mp3_roles_path.exists():
        raise AvalonBGError("Mp3 file doesn't exist for this game!")

    return mp3_roles_path.as_posix()
